"""sec-value

Revision ID: f7b416c59e27
Revises: 0a4bbaa82cd1
Create Date: 2025-11-01 15:30:24.903828

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "f7b416c59e27"
down_revision: Union[str, None] = "0a4bbaa82cd1"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("sec", schema=None) as batch_op:
        batch_op.add_column(sa.Column("value", sa.Float(), nullable=True))

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("sec", schema=None) as batch_op:
        batch_op.drop_column("value")

    # ### end Alembic commands ###



from typing import Annotated, List, Any, Callable

from pydantic import BaseModel, model_validator, BeforeValidator, Field

from bearish.models.assets.assets import Assets
from bearish.models.base import BaseAssets, Ticker
from bearish.utils.utils import remove_duplicates, remove_duplicates_string


class BaseAssetQuery(BaseModel):
    @model_validator(mode="after")
    def validate_query(self) -> Any:
        if all(not getattr(self, field) for field in self.model_fields):
            raise ValueError("At least one query parameter must be provided")
        return self


class Symbols(BaseAssets):
    equities: Annotated[
        List[Ticker], BeforeValidator(remove_duplicates), Field(default_factory=list)
    ]
    etfs: Annotated[
        List[Ticker], BeforeValidator(remove_duplicates), Field(default_factory=list)
    ]
    currencies: Annotated[
        List[Ticker], BeforeValidator(remove_duplicates), Field(default_factory=list)
    ]
    cryptos: Annotated[
        List[Ticker], BeforeValidator(remove_duplicates), Field(default_factory=list)
    ]
    index: Annotated[
        List[Ticker], BeforeValidator(remove_duplicates), Field(default_factory=list)
    ]

    def filter(self, func: Callable[[Ticker], Any]) -> "Symbols":
        equities = [e for e in self.equities if func(e)]
        etfs = [e for e in self.etfs if func(e)]
        currencies = [e for e in self.currencies if func(e)]
        cryptos = [e for e in self.cryptos if func(e)]
        index = [e for e in self.index if func(e)]
        return Symbols(
            equities=equities,
            etfs=etfs,
            currencies=currencies,
            cryptos=cryptos,
            index=index,
        )

    def equities_symbols(self) -> List[str]:
        return [t.symbol for t in self.equities]

    def etfs_symbols(self) -> List[str]:
        return [t.symbol for t in self.etfs]

    def empty(self) -> bool:
        return not any(
            [
                self.equities,
                self.etfs,
                self.currencies,
                self.cryptos,
                self.index,
            ]
        )

    def all(self) -> List[str]:
        return [
            t.symbol
            for t in self.equities
            + self.etfs
            + self.currencies
            + self.cryptos
            + self.index
        ]


class AssetQuery(BaseAssetQuery):
    countries: Annotated[
        List[str],
        BeforeValidator(remove_duplicates_string),
        Field(default_factory=list),
    ]
    exchanges: Annotated[
        List[str],
        BeforeValidator(remove_duplicates_string),
        Field(default_factory=list),
    ]
    excluded_sources: Annotated[
        List[str],
        BeforeValidator(remove_duplicates_string),
        Field(default_factory=list),
    ]
    symbols: Symbols = Field(default=Symbols())  # type: ignore

    def update_symbols(self, assets: Assets) -> None:
        for field in assets.model_fields:
            if field == "failed_query":
                continue
            symbols = sorted(
                {
                    Ticker(symbol=asset.symbol, exchange=asset.exchange)
                    for asset in getattr(assets, field)
                }
                | set(getattr(self.symbols, field)),
                key=lambda x: x.symbol,
            )
            setattr(
                self.symbols,
                field,
                symbols,
            )



class InvalidApiKeyError(Exception):
    pass


class LimitApiKeyReachedError(Exception):
    pass


class TickerNotFoundError(Exception):
    pass


class FinancialDataError(Exception):
    pass


class IncompleteDataError(Exception):
    pass



"""sec

Revision ID: 0a4bbaa82cd1
Revises: ca2d5f47456a
Create Date: 2025-10-28 09:22:18.123782

"""

from typing import Sequence, Union

import sqlmodel
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "0a4bbaa82cd1"
down_revision: Union[str, None] = "ca2d5f47456a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "sec",
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("cusip", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("shares", sa.Integer(), nullable=True),
        sa.Column("period", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("filed_date", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("ticker", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("company_name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.PrimaryKeyConstraint("name", "period", "filed_date", "source"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("sec", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_sec_filed_date"), ["filed_date"], unique=False
        )
        batch_op.create_index(batch_op.f("ix_sec_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_sec_period"), ["period"], unique=False)
        batch_op.create_index(batch_op.f("ix_sec_source"), ["source"], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("sec", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_sec_source"))
        batch_op.drop_index(batch_op.f("ix_sec_period"))
        batch_op.drop_index(batch_op.f("ix_sec_name"))
        batch_op.drop_index(batch_op.f("ix_sec_filed_date"))

    op.drop_table("sec")
    # ### end Alembic commands ###



import abc
import datetime
from datetime import date
from typing import Dict, Any, ClassVar, Optional

from pydantic import (
    BaseModel,
    ConfigDict,
    model_validator,
    field_validator,
    Field,
)

from bearish.types import Sources


class BaseAssets(BaseModel):
    equities: Any
    cryptos: Any
    etfs: Any
    currencies: Any
    index: Any


class Ticker(BaseModel):
    symbol: str
    exchange: Optional[str] = None
    source: Optional[Sources] = None

    def __hash__(self) -> int:
        return hash(self.symbol)


class SourceBase(BaseModel, abc.ABC):
    __source__: Sources
    __alias__: ClassVar[Dict[str, str]] = {}
    __api_key__: ClassVar[str]
    model_config = ConfigDict(populate_by_name=True, extra="forbid")


class BaseTrackerQuery(BaseModel):
    exchange: Optional[str] = None


class TrackerQuery(BaseTrackerQuery):
    reference_date: Optional[datetime.date] = None
    delay: int = 5


class BaseTracker(BaseTrackerQuery):
    source: str
    symbol: str
    date: datetime.date = Field(default_factory=lambda: date(1970, 1, 1))


class PriceTracker(BaseTracker): ...


class FinancialsTracker(BaseTracker): ...


class DataSourceBase(SourceBase, Ticker):
    source: Sources
    date: datetime.date
    created_at: datetime.date

    @field_validator("date", mode="before")
    def _date_validator(cls, value: str | datetime.date) -> datetime.date:  # noqa: N805
        if isinstance(value, str):
            return datetime.datetime.strptime(value, "%Y-%m-%d").date()
        return value

    @model_validator(mode="before")
    def _validate(cls, metrics: Dict[str, Any]) -> Dict[str, Any]:  # noqa: N805
        if not hasattr(cls, "__source__") and "source" not in metrics:
            raise ValueError("No source specified for financial metrics")
        source = cls.__source__ if hasattr(cls, "__source__") else metrics["source"]
        default_keys = {field: field for field in cls.model_fields}
        default_keys.pop("date", None)
        alias = cls.__alias__.copy()
        alias = {**alias, **default_keys}

        created_at = date.today()
        current_date = metrics.get("date", created_at)
        if isinstance(current_date, datetime.datetime):
            current_date = current_date.date()
        return (
            {"date": current_date}
            | {
                alias.get(key, key): value
                for key, value in metrics.items()
                if key in alias
            }
            | {
                "created_at": created_at,
                "source": source,
            }
        )



"""index table

Revision ID: 79d40a2eeef5
Revises: b62d0afd2675
Create Date: 2025-10-26 13:03:29.683677

"""

from typing import Sequence, Union

import sqlmodel
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "79d40a2eeef5"
down_revision: Union[str, None] = "b62d0afd2675"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "index",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("base_symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("modifier", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("summary", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("currency", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("country", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("category_group", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("category", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("index", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_index_country"), ["country"], unique=False)
        batch_op.create_index(batch_op.f("ix_index_source"), ["source"], unique=False)
        batch_op.create_index(batch_op.f("ix_index_symbol"), ["symbol"], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("index", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_index_symbol"))
        batch_op.drop_index(batch_op.f("ix_index_source"))
        batch_op.drop_index(batch_op.f("ix_index_country"))

    op.drop_table("index")
    # ### end Alembic commands ###



import abc
import logging
from functools import wraps
from io import StringIO
from typing import List, Optional, Type, Callable, Any, cast

import pandas as pd
import requests  # type: ignore
from pydantic import ConfigDict, validate_call, BaseModel, Field

from bearish.exceptions import InvalidApiKeyError, LimitApiKeyReachedError
from bearish.exchanges.exchanges import Countries, Exchanges, exchanges_factory
from bearish.models.query.query import AssetQuery
from bearish.models.assets.assets import Assets
from bearish.models.base import SourceBase, DataSourceBase, Ticker

from bearish.models.financials.base import Financials
from bearish.models.price.price import Price
from bearish.types import Sources, SeriesLength
from bearish.utils.utils import observability

logger = logging.getLogger(__name__)


class ValidTickers(BaseModel):
    sources: List[Sources] = Field(default_factory=list)
    exchanges: List[str] = Field(default_factory=list)

    def is_valid(self, ticker: Ticker) -> bool:
        if not self.exchanges and not self.sources:
            return True
        elif self.exchanges and not self.sources:
            return ticker.exchange in self.exchanges
        elif self.sources and not self.exchanges:
            return ticker.source in self.sources
        else:
            return bool(
                ticker.exchange is not None
                and self.exchanges
                and ticker.source in self.sources
            )


class ApiUsage(BaseModel):
    calls: int = 0
    calls_limit: Optional[int] = None

    def limit_reached(self) -> bool:
        return self.calls >= self.calls_limit if self.calls_limit else False

    def add_api_calls(self, calls: int) -> None:
        self.calls += calls


def check_api_limit(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(self: "AbstractSource", *args: Any, **kwargs: Any) -> Any:
        if self.api_usage.limit_reached():
            raise LimitApiKeyReachedError(
                f"API '{self.__source__}' Limit reached : {self.api_usage.calls_limit}"
            )
        logger.debug(f"API '{self.__source__}' call count: {self.api_usage.calls}")
        return func(self, *args, **kwargs)

    return cast(Callable[..., Any], wrapper)


class AbstractSource(SourceBase, abc.ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    countries: List[Countries]
    exchanges: Exchanges = Field(default_factory=exchanges_factory)
    api_usage: ApiUsage = Field(default_factory=ApiUsage)
    pause: int = 60

    @validate_call(validate_return=True)
    @check_api_limit
    @observability
    def read_assets(self, query: Optional[AssetQuery] = None) -> Assets:
        query_ = None
        if query:
            query_ = self.exchanges.get_asset_query(query, self.countries)
            logger.debug(
                f"Reading assets from {type(self).__name__} with {len(query_.symbols.all())} tickers"
            )
        try:
            return self._read_assets(query_)
        except Exception as e:
            logger.error(
                f"Error reading assets from {type(self).__name__}: {e}", exc_info=True
            )
            return Assets()

    @validate_call(validate_return=True)
    @check_api_limit
    @observability
    def read_financials(self, tickers: List[Ticker]) -> List[Financials]:

        tickers = [
            ticker
            for ticker in tickers
            if self.exchanges.ticker_belongs_to_countries(
                ticker, countries=self.countries
            )
        ]

        try:
            logger.info(f"Reading Financials from {type(self).__name__}")
            return self._read_financials([t.symbol for t in tickers])

        except InvalidApiKeyError as e:
            raise e
        except Exception as e:
            logger.error(f"Error reading Financials from {type(self).__name__}: {e}")

        return []

    @validate_call(validate_return=True)
    @check_api_limit
    @observability
    def read_series(
        self, tickers: List[Ticker], type_: SeriesLength, apply_filter: bool = True
    ) -> List[Price]:
        if apply_filter:
            tickers = [
                ticker
                for ticker in tickers
                if self.exchanges.ticker_belongs_to_countries(
                    ticker, countries=self.countries
                )
            ]
        try:
            prices_ = self._read_series([t.symbol for t in tickers], type_)
            return [p for p in prices_ if p.valid()]
        except InvalidApiKeyError as e:
            raise e
        except Exception as e:
            logger.error(f"Error reading prices from {type(self).__name__}: {e}")

        return []

    @abc.abstractmethod
    def _read_financials(self, tickers: List[str]) -> List[Financials]: ...

    @abc.abstractmethod
    def _read_assets(self, query: Optional[AssetQuery] = None) -> Assets: ...

    @abc.abstractmethod
    def _read_series(self, tickers: List[str], type: SeriesLength) -> List[Price]: ...

    @abc.abstractmethod
    def set_api_key(self, api_key: str) -> None: ...

    def set_pause(self, value: int) -> None:
        self.pause = value

    def __hash__(self) -> int:
        return hash(self.__source__)


class UrlSource(BaseModel):
    url: str
    results: List[SourceBase] = Field(default_factory=list)
    type_class: Type[SourceBase]
    filters: Optional[List[str]] = None
    renames: Optional[dict[str, str]] = None


class UrlSources(BaseModel):
    equity: UrlSource
    crypto: UrlSource
    currency: Optional[UrlSource] = Field(None)
    etf: UrlSource
    index: Optional[UrlSource] = Field(None)

    def to_assets(self) -> Assets:

        return Assets(
            equities=self.equity.results,
            cryptos=self.crypto.results,
            currencies=self.currency.results if self.currency else [],
            etfs=self.etf.results,
            index=self.index.results if self.index else [],
        )


class DatabaseCsvSource(AbstractSource):
    __url_sources__: UrlSources

    def set_api_key(self, api_key: str) -> None: ...

    def _read_assets(self, query: Optional[AssetQuery] = None) -> Assets:
        sources = self.__url_sources__
        for field in sources.model_fields:
            try:
                url_source = getattr(sources, field)
                if url_source is None:
                    continue
                response = requests.get(url_source.url, timeout=10)
                if not response.ok:
                    raise Exception(f"Failed to download data from {url_source.url}")
                data = pd.read_csv(StringIO(response.text))
                url_source.results = self._from_dataframe(
                    data, url_source.type_class, url_source.filters, url_source.renames
                )
            except Exception as e:
                logger.error(f"Failed to download data from {url_source.url}: {e}")
                continue
        return sources.to_assets()

    def _from_dataframe(
        self,
        data: pd.DataFrame,
        databaseclass: Type[DataSourceBase],
        filters: Optional[list[str]] = None,
        renames: Optional[dict[str, str]] = None,
    ) -> List[DataSourceBase]:
        if renames:
            data = data.rename(columns=renames)
        if filters:
            data = data.dropna(subset=filters)

        equities_mapping = [equity.to_dict() for _, equity in data.iterrows()]
        return [databaseclass(**equity_mapping) for equity_mapping in equities_mapping]

    def _read_financials(self, tickers: List[str]) -> List[Financials]:
        return [Financials()]

    def _read_series(self, tickers: List[str], type: str) -> List[Price]:
        return []



import re
import tempfile
import xml.etree.ElementTree as ET
from collections import defaultdict
from datetime import date, timedelta
from io import StringIO
from pathlib import Path
from typing import Optional, List, Dict, TYPE_CHECKING, Self

import pandas as pd
import requests  # type: ignore
from pydantic import BaseModel
from sec_cik_mapper import StockMapper  # type: ignore
from sec_edgar_downloader import Downloader  # type: ignore
import yfinance as yf  # type: ignore

if TYPE_CHECKING:
    from bearish.database.crud import BearishDb

CIKS = [
    "0001081316",
    "0001274791",
    "0000829771",
    "0001067983",
    "0000109694",
    "0001015867",
    "0000895421",
    "0001364742",
    "0000880052",
    "0000102909",
    "0000735286",
    "0000093751",
    "0000315066",
    "0000080255",
    "0000017283",
    "0001562230",
    "0001214717",
    "0001009268",
    "0001179392",
    "0001037389",
    "0001350694",
    "0001423053",
    "0001167483",
    "0001336528",
    "0000921669",
    "0001061768",
    "0001103804",
    "0001603466",
    "0000070858",
]


def normalize(s: str) -> str:
    s = s.upper()
    s = s.replace("COMMON STOCK", "")  # uppercase
    s = s.replace("COMMON SHARES", "")  # uppercase
    s = s.replace("CLASS A ORDINARY SHARES", "")  # uppercase
    s = s.replace("CLASS A COMMON STOCK", "")  # uppercase
    s = s.replace("INC", "")  # uppercase
    return re.sub(r"[^A-Z0-9]", "", s)


def company_ticker() -> Dict[str, str]:
    nasdaq_url = "https://www.nasdaqtrader.com/dynamic/symdir/nasdaqlisted.txt"
    other_url = "https://www.nasdaqtrader.com/dynamic/symdir/otherlisted.txt"
    data_ = []
    headers = {"User-Agent": "YourName Contact@Email.com", "Accept": "text/plain"}
    mapper = StockMapper()

    for url in [nasdaq_url, other_url]:
        response = requests.get(url, headers=headers, timeout=30)
        response.raise_for_status()
        data = pd.read_csv(StringIO(response.text), delimiter="|")
        data = data.rename(columns={"ACT Symbol": "Symbol"})
        data_.append(data)
    data_all = pd.concat(data_, ignore_index=True)
    ticker_mapping = {
        normalize(str(e["Security Name"])): e["Symbol"]
        for e in data_all[["Symbol", "Security Name"]].to_dict(orient="records")
    }
    ticker_mapping.update(
        {normalize(n): t for t, n in mapper.ticker_to_company_name.items()}  # type: ignore
    )
    return ticker_mapping


TICKER_MAPPING = company_ticker()


class Sec(BaseModel):
    name: Optional[str] = None
    cusip: Optional[str] = None
    shares: Optional[int] = None
    period: Optional[str] = None
    filed_date: Optional[str] = None
    ticker: Optional[str] = None
    source: Optional[str] = None
    company_name: Optional[str] = None
    value: Optional[float] = None

    def __hash__(self) -> int:
        return hash((self.name, self.source, self.period, self.filed_date, self.cusip))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Sec):
            return hash(self) == hash(other)
        return False

    def set_share(self, shares: int) -> Self:
        self.shares = shares
        return self


def _info_table_to_rows(
    sec_text: str, source: str, ticker_mapping: Optional[Dict[str, str]] = None
) -> List[Sec]:
    ticker_mapping = ticker_mapping or {}
    xml = re.search(  # type: ignore
        r"<DOCUMENT>.*?<TYPE>\s*INFORMATION TABLE.*?<XML>(.*?)</XML>.*?</DOCUMENT>",
        sec_text,
        re.DOTALL | re.IGNORECASE,
    ).group(1)
    ns = {"it": "http://www.sec.gov/edgar/document/thirteenf/informationtable"}
    root = ET.fromstring(xml.lstrip())  # noqa: S314
    period = re.search(r"CONFORMED PERIOD OF REPORT:\s+(\d{8})", sec_text)
    filed_date_ = re.search(r"FILED AS OF DATE:\s+(\d{8})", sec_text)
    company_name_ = re.search(r"COMPANY CONFORMED NAME:\s*([^\r\n]+)", sec_text)

    if not period:
        raise ValueError("period is required")
    per_dt = period.group(1)
    period_report = f"{per_dt[0:4]}-{per_dt[4:6]}-{per_dt[6:8]}"

    if not filed_date_:
        raise ValueError("Filed date is missing")
    filed_date__ = filed_date_.group(1)
    filed_date = f"{filed_date__[0:4]}-{filed_date__[4:6]}-{filed_date__[6:8]}"

    company_name = None
    if company_name_:
        company_name = company_name_.group(1)

    rows = []
    for t in root.findall(".//it:infoTable", ns):
        name = t.findtext("it:nameOfIssuer", default="", namespaces=ns).strip()
        cusip = t.findtext("it:cusip", default="", namespaces=ns).strip()
        # shares live inside shrsOrPrnAmt/sshPrnamt
        shares_text = t.findtext(
            "it:shrsOrPrnAmt/it:sshPrnamt", default="0", namespaces=ns
        )
        shares = int(float(shares_text))
        rows.append(
            Sec.model_validate(
                {
                    "name": name,
                    "cusip": cusip,
                    "shares": shares,
                    "period": period_report,
                    "source": source,
                    "filed_date": filed_date,
                    "ticker": ticker_mapping.get(normalize(name)),
                    "company_name": company_name,
                }
            )
        )
    return rows


class Secs(BaseModel):
    secs: List[Sec]

    def write(self, bearish_db: "BearishDb") -> None:
        bearish_db.write_sec(self.secs)

    @classmethod
    def upload(cls, bearish_db: "BearishDb", date_: Optional[date] = None) -> None:
        for cik in CIKS:
            sec = cls.from_sec_13f_hr(cik, date_=date_)
            sec.write(bearish_db)
        cls.update_values(bearish_db)

    @classmethod
    def update_values(cls, bearish_db: "BearishDb") -> None:
        prices = {}
        tickers = bearish_db.read_query(
            query="SELECT DISTINCT ticker from sec WHERE ticker NOT NULL"
        )
        tickers_ = tickers["ticker"].tolist()
        batch_size = 100
        for i in range(0, len(tickers_), batch_size):
            batch = tickers_[i : i + batch_size]
            data = yf.download(batch, period="5d", auto_adjust=True, timeout=60)
            prices.update(data["Close"].dropna(axis=1).iloc[-1].to_dict())
        for symbol, price in prices.items():
            secs = bearish_db.read_sec(symbol)
            for sec in secs:
                sec.value = price * sec.shares
            bearish_db.write_sec(secs)

    @classmethod
    def from_sec_13f_hr(cls, source: str, date_: Optional[date] = None) -> "Secs":
        rows = []
        unique_secs = []
        groups = defaultdict(list)
        date__ = date_ or date.today() - timedelta(days=31 * 4)

        with tempfile.TemporaryDirectory() as tmpdir:
            dl = Downloader(
                "MyCompanyName", "email@example.com", download_folder=tmpdir
            )
            dl.get("13F-HR", source, after=date__.strftime("%Y-%m-%d"))
            for p in list(Path(tmpdir).rglob("*.txt")):
                rows.extend(
                    _info_table_to_rows(
                        p.read_text(), source, ticker_mapping=TICKER_MAPPING
                    )
                )
            for item in rows:
                groups[item].append(item)
            for key, values in groups.items():
                shares = sum([s.shares for s in values if s.shares is not None])
                unique_secs.append(key.set_share(shares))
        return cls(secs=unique_secs)



from typing import Annotated, Optional, Dict

from pydantic import BeforeValidator, Field

from bearish.models.assets.base import BaseComponent
from bearish.utils.utils import to_string, to_float


class Etf(BaseComponent):
    category_group: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(default=None),
    ]
    family: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(default=None),
    ]

    three_year_earnings_growth: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Earnings growth over the past three years."),
    ]
    annual_holdings_turnover: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Percentage of portfolio holdings replaced annually."),
    ]
    annual_report_expense_ratio: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            None,
            description="Annual cost of managing the fund as a percentage of total assets.",
        ),
    ]
    holding_percent: Optional[Dict[str, float]] = Field(
        None, description="Percentage of the ETF allocated to specific holdings."
    )
    median_market_cap: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Median market capitalization of the ETF's holdings."),
    ]
    price_to_book: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Price-to-book ratio of the ETF."),
    ]
    price_to_cashflow: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Price-to-cashflow ratio of the ETF."),
    ]
    price_to_earnings: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Price-to-earnings ratio of the ETF."),
    ]
    price_to_sales: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Price-to-sales ratio of the ETF."),
    ]
    sector_weightings: Optional[Dict[str, float]] = Field(
        None, description="Weightings of the ETF across various sectors."
    )
    total_net_assets: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Total net assets managed by the ETF."),
    ]

    beta_3_year: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Beta value of the ETF over the past three years."),
    ]

    category: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(None, description="Category of the ETF (e.g., 'Large Blend')."),
    ]
    fund_family: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(None, description="Fund family managing the ETF."),
    ]
    fund_inception_date: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Inception date of the ETF in epoch format."),
    ]
    legal_type: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            None,
            description="Legal classification of the ETF (e.g., 'Exchange Traded Fund').",
        ),
    ]
    long_business_summary: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            None, description="Detailed description of the ETF's investment objective."
        ),
    ]
    long_name: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(None, description="Full name of the ETF."),
    ]
    nav_price: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Net asset value (NAV) price of the ETF."),
    ]
    quote_type: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(None, description="Type of the security (e.g., 'ETF')."),
    ]
    short_name: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(None, description="Shortened name of the ETF."),
    ]
    three_year_average_return: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Three-year average return of the ETF."),
    ]
    trailing_annual_dividend_rate: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Annual dividend rate paid by the ETF."),
    ]
    trailing_annual_dividend_yield: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Annual dividend yield of the ETF."),
    ]
    trailing_pe: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Trailing price-to-earnings ratio of the ETF."),
    ]
    yield_: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Yield of the ETF."),
    ]
    ytd_return: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(None, description="Year-to-date return of the ETF."),
    ]



import functools
import logging
import time
from datetime import datetime, timedelta
from math import isnan
from typing import Any, Optional, List, Dict, Callable

import pandas as pd

from bearish.models.base import Ticker
from bearish.types import SeriesLength

logger = logging.getLogger(__name__)


def to_float(value: Any) -> Optional[float]:
    if value == "None":
        return None
    if value is None:
        return None
    if isinstance(value, str):
        try:
            return float(value)
        except ValueError:
            return None
    return float(value)


def to_datetime(value: Any) -> datetime:
    if isinstance(value, str):
        return datetime.strptime(value, "%Y-%m-%d")
    elif isinstance(value, pd.Timestamp):
        if value.tz is not None:
            value = value.tz_convert(None)
        return value.to_pydatetime()  # type: ignore
    elif isinstance(value, datetime):
        return value
    else:
        raise ValueError(f"Invalid datetime value: {value}")


def to_string(value: Any) -> Optional[str]:
    if value is None or (isinstance(value, float) and isnan(value)):
        return None
    if value == "None":
        return None
    return str(value)


def format_capitalize(value: Any) -> Optional[str]:
    country = to_string(value)
    if country is None:
        return None
    return country.capitalize()


def remove_duplicates(value: list[Ticker]) -> list[Ticker]:
    if not value:
        return []
    tickers = [Ticker.model_validate(t) for t in value]
    return list({t.symbol: t for t in tickers}.values())


def remove_duplicates_string(value: list[str]) -> list[str]:
    if not value:
        return []
    return list(set(value))


def get_start_date(type: SeriesLength) -> Optional[str]:
    from_ = None
    if type != "max":
        past_date = datetime.today() - timedelta(days=int(type.replace("d", "")))
        from_ = str(past_date.strftime("%Y-%m-%d"))
    return from_


def to_dataframe(datas: List[Any]) -> pd.DataFrame:
    data = pd.DataFrame.from_records([p.model_dump() for p in datas])
    if data.empty:
        return data
    data = data.set_index("date", inplace=False)
    data = data.sort_index(inplace=False)

    data.index = pd.to_datetime(data.index, utc=True)
    data = data[~data.index.duplicated(keep="first")]
    return data


def batch(objects: List[Any], size: int) -> List[List[Any]]:
    return [objects[i : i + size] for i in range(0, len(objects), size)]


def safe_get(data: Dict[str, Any], attribute: str) -> Dict[str, Any]:
    if not isinstance(data, dict):
        logger.warning(
            f"Expected a dictionary, got {type(data)} instead with value {data}"
        )
        return {}
    value = data.get(attribute, {})
    return value if isinstance(value, dict) else {}


def observability(func: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            start = time.perf_counter()
            result = func(*args, **kwargs)
            time_elapsed = time.perf_counter() - start
            logger.debug("=========================================================")
            logger.debug(f"Function {func.__name__} took {time_elapsed:.2f} seconds")
            logger.debug("=========================================================")
            return result
        except Exception as e:
            logger.error(e, exc_info=True, stack_info=True)
            raise e

    return wrapper



from typing import Optional

from pydantic import Field

from bearish.models.base import DataSourceBase


class FinancialMetrics(DataSourceBase):

    ebitda: Optional[float] = Field(
        None,
        description="Earnings before interest, taxes, depreciation, and amortization",
    )
    net_income: Optional[float] = Field(
        None, description="Total earnings or profit of the company"
    )
    pe_ratio: Optional[float] = Field(None)
    market_capitalization: Optional[float] = Field(None)
    basic_eps: Optional[float] = Field(
        None,
        description="Earnings per share for basic shares outstanding",
    )
    diluted_eps: Optional[float] = Field(
        None, alias="Diluted EPS", description="Earnings per share for diluted shares"
    )
    total_revenue: Optional[float] = Field(
        None,
        description="Total revenue generated by the company",
    )
    operating_revenue: Optional[float] = Field(
        None, description="Revenue from operating activities"
    )
    gross_profit: Optional[float] = Field(
        None,
        description="Revenue after deducting cost of goods sold",
    )
    total_expenses: Optional[float] = Field(
        None,
        description="Total expenses incurred by the company",
    )
    operating_income: Optional[float] = Field(
        None, description="Income generated from operations"
    )
    profit_margin: Optional[float] = Field(None, description="Net profit margin")
    cost_of_revenue: Optional[float] = Field(
        None, description="Cost incurred to generate revenue"
    )
    tax_provision: Optional[float] = Field(None, description="Provision for income tax")
    tax_rate: Optional[float] = Field(
        None, description="Effective tax rate used in calculations"
    )


class QuarterlyFinancialMetrics(FinancialMetrics): ...



from typing import Annotated, Optional

from pydantic import BeforeValidator, Field

from bearish.models.assets.base import BaseComponent
from bearish.models.base import Ticker
from bearish.utils.utils import to_string

PRICE_INDEX = [
    Ticker(symbol="^GSPC"),  # S&P 500
    Ticker(symbol="^DJI"),  # Dow Jones Industrial Average
    Ticker(symbol="^IXIC"),  # Nasdaq Composite
    Ticker(symbol="^RUT"),  # Russell 2000
    Ticker(symbol="SPG1200"),  # S&P Global 1200 (varies by feed)
    Ticker(symbol="URTH"),  # MSCI World (ETF proxy)
    Ticker(symbol="IOO"),  # S&P Global 100 (ETF proxy)
    Ticker(symbol="SPBMISR"),  # S&P Global BMI (varies)
    Ticker(symbol="^FTSE"),  # FTSE 100
    Ticker(symbol="MCX"),  # FTSE 250 (LSE)
    Ticker(symbol="^FCHI"),  # CAC 40
    Ticker(symbol="^GDAXI"),  # DAX
    Ticker(symbol="^N225"),  # Nikkei 225
    Ticker(symbol="^TOPX"),  # TOPIX (provider dependent)
    Ticker(symbol="^HSI"),  # Hang Seng
    Ticker(symbol="000001.SS"),  # Shanghai Composite
    Ticker(symbol="^AXJO"),  # S&P/ASX 200
    Ticker(symbol="^GSPTSE"),  # TSX Composite
    Ticker(symbol="^BSESN"),  # BSE Sensex
    Ticker(symbol="^NSEI"),  # Nifty 50
    Ticker(symbol="^KS11"),  # KOSPI
    Ticker(symbol="^SET.BK"),  # SET Thailand
    Ticker(symbol="^BVSP"),  # Ibovespa
    Ticker(symbol="^TX60"),  # S&P/TSX 60 (varies)
    Ticker(symbol="FTSEMIB.MI"),  # FTSE MIB
    Ticker(symbol="^SSMI"),  # Swiss Market Index
    Ticker(symbol="^STOXX50E"),  # Euro Stoxx 50
    Ticker(symbol="^STOXX"),  # Stoxx Europe 600
    Ticker(symbol="EEM"),  # MSCI Emerging Markets (ETF proxy)
    Ticker(symbol="ACWI"),  # MSCI ACWI (ETF proxy)
    Ticker(symbol="^SPTSE"),  # S&P/TSX Composite
    Ticker(symbol="^AXKO"),  # S&P/ASX 50 (varies)
    Ticker(symbol="ILF"),  # S&P Latin America 40 (ETF proxy)
    Ticker(symbol="AIA"),  # S&P Asia 50 (ETF proxy)
    Ticker(symbol="TPX100"),  # S&P/TOPIX 150 (varies)
    Ticker(symbol="^DJGT50"),  # DJ Global Titans 50 (varies)
    Ticker(symbol="SP1500"),  # S&P Composite 1500 (varies)
    Ticker(symbol="^RUA"),  # Russell 3000
    Ticker(symbol="399001.SZ"),  # SZSE Component
    Ticker(symbol="000300.SS"),  # CSI 300
    Ticker(symbol="^FTAS"),  # FTSE All-Share
    Ticker(symbol="IEUR"),  # MSCI Europe (ETF proxy)
    Ticker(symbol="IPAC"),  # MSCI Asia Pacific (ETF proxy)
    Ticker(symbol="FM"),  # MSCI Frontier Markets (ETF proxy)
    Ticker(symbol="^MXX"),  # S&P/BMV IPC Mexico
    Ticker(symbol="^AEX"),  # AEX
    Ticker(symbol="^OMXC20"),  # OMX Copenhagen 20
    Ticker(symbol="^BFX"),  # BEL 20
    Ticker(symbol="XU100.IS"),  # BIST 100 Turkey
    Ticker(symbol="PSEI.PS"),  # PSEi Philippines
    Ticker(symbol="0020.Z"),  # PSEi Philippines
]


class Index(BaseComponent):
    category_group: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(default=None),
    ]
    category: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(None),
    ]



from typing import List

from pydantic import Field, BaseModel

from bearish.exceptions import TickerNotFoundError
from bearish.models.assets.crypto import Crypto
from bearish.models.assets.currency import Currency
from bearish.models.assets.equity import Equity
from bearish.models.assets.etfs import Etf
from bearish.models.assets.index import Index
from bearish.models.base import BaseAssets, Ticker


class FailedQueryAssets(BaseModel):
    symbols: List[Ticker] = Field(default_factory=list)


class Assets(BaseAssets):
    equities: List[Equity] = Field(default_factory=list)
    cryptos: List[Crypto] = Field(default_factory=list)
    etfs: List[Etf] = Field(default_factory=list)
    currencies: List[Currency] = Field(default_factory=list)
    index: List[Index] = Field(default_factory=list)
    failed_query: FailedQueryAssets = Field(default_factory=FailedQueryAssets)

    def get_one_equity(self) -> Equity:
        if not self.equities:
            raise TickerNotFoundError("No equities found")
        return self.equities[0]

    def is_empty(self) -> bool:
        return not any(
            [self.equities, self.cryptos, self.etfs, self.currencies, self.index]
        )

    def add(self, assets: "Assets") -> None:
        self.equities.extend(assets.equities)
        self.cryptos.extend(assets.cryptos)
        self.etfs.extend(assets.etfs)
        self.currencies.extend(assets.currencies)
        self.index.extend(assets.index)

    def symbols(self) -> List[Ticker]:
        return [
            Ticker(symbol=asset.symbol, source=asset.source, exchange=asset.exchange)
            for asset in self.equities
            + self.cryptos
            + self.etfs
            + self.currencies
            + self.index
        ]



"""price index

Revision ID: ca2d5f47456a
Revises: 79d40a2eeef5
Create Date: 2025-10-27 06:10:22.467752

"""

from typing import Sequence, Union

import sqlmodel
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "ca2d5f47456a"
down_revision: Union[str, None] = "79d40a2eeef5"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "priceindex",
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("open", sa.Float(), nullable=False),
        sa.Column("high", sa.Float(), nullable=False),
        sa.Column("low", sa.Float(), nullable=False),
        sa.Column("close", sa.Float(), nullable=False),
        sa.Column("volume", sa.Float(), nullable=False),
        sa.Column("dividends", sa.Float(), nullable=True),
        sa.Column("stock_splits", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source", "date"),
    )
    with op.batch_alter_table("priceindex", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_priceindex_date"), ["date"], unique=False)
        batch_op.create_index(
            batch_op.f("ix_priceindex_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_priceindex_symbol"), ["symbol"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("priceindex", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_priceindex_symbol"))
        batch_op.drop_index(batch_op.f("ix_priceindex_source"))
        batch_op.drop_index(batch_op.f("ix_priceindex_date"))

    op.drop_table("priceindex")
    # ### end Alembic commands ###



from typing import ClassVar, Dict, List

from pydantic import Field

from bearish.exchanges.exchanges import Countries
from bearish.models.assets.crypto import Crypto
from bearish.models.assets.currency import Currency
from bearish.models.assets.equity import Equity
from bearish.models.assets.etfs import Etf
from bearish.models.assets.index import Index
from bearish.models.base import SourceBase
from bearish.sources.base import (
    UrlSources,
    UrlSource,
    DatabaseCsvSource,
)
from bearish.types import Sources

RAW_EQUITIES_DATA_URL = "https://raw.githubusercontent.com/JerBouma/FinanceDatabase/refs/heads/main/database/equities.csv"
RAW_CRYPTO_DATA_URL = "https://raw.githubusercontent.com/JerBouma/FinanceDatabase/refs/heads/main/database/cryptos.csv"
RAW_CURRENCY_DATA_URL = "https://raw.githubusercontent.com/JerBouma/FinanceDatabase/refs/heads/main/database/currencies.csv"
RAW_ETF_DATA_URL = "https://raw.githubusercontent.com/JerBouma/FinanceDatabase/refs/heads/main/database/etfs.csv"
RAW_INDEX_DATA_URL = "https://raw.githubusercontent.com/JerBouma/FinanceDatabase/refs/heads/main/database/indices.csv"


class FinanceDatabaseBase(SourceBase):
    __source__: Sources = "FinanceDatabase"
    __alias__: ClassVar[Dict[str, str]] = {
        "symbol": "symbol",
        "cryptocurrency": "cryptocurrency",
    }


class FinanceDatabaseEquity(FinanceDatabaseBase, Equity): ...


class FinanceDatabaseCrypto(FinanceDatabaseBase, Crypto): ...


class FinanceDatabaseCurrency(FinanceDatabaseBase, Currency): ...


class FinanceDatabaseEtf(FinanceDatabaseBase, Etf): ...


class FinanceDatabaseIndex(FinanceDatabaseBase, Index): ...


class FinanceDatabaseSource(FinanceDatabaseBase, DatabaseCsvSource):
    countries: List[Countries] = Field(default_factory=list)
    __url_sources__ = UrlSources(
        equity=UrlSource(
            url=RAW_EQUITIES_DATA_URL,
            type_class=FinanceDatabaseEquity,
            filters=["symbol", "country"],
            renames={"Unnamed: 0": "symbol"},
        ),
        crypto=UrlSource(
            url=RAW_CRYPTO_DATA_URL,
            type_class=FinanceDatabaseCrypto,
            filters=["symbol", "cryptocurrency"],
            renames={"Unnamed: 0": "symbol"},
        ),
        currency=UrlSource(
            url=RAW_CURRENCY_DATA_URL,
            type_class=FinanceDatabaseCurrency,
            filters=["symbol"],
            renames={"Unnamed: 0": "symbol"},
        ),
        etf=UrlSource(
            url=RAW_ETF_DATA_URL,
            type_class=FinanceDatabaseEtf,
            filters=["symbol"],
            renames={"Unnamed: 0": "symbol"},
        ),
        index=UrlSource(
            url=RAW_INDEX_DATA_URL,
            type_class=FinanceDatabaseIndex,
            filters=["symbol"],
            renames={"Unnamed: 0": "symbol"},
        ),
    )



from typing import Annotated, Optional

from pydantic import BeforeValidator, Field

from bearish.models.assets.base import BaseComponent
from bearish.utils.utils import to_string, to_float


class BaseEquity(BaseComponent):
    sector: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            None,
            description="Broad sector to which the company belongs, such as 'Real Estate' or 'Technology'",
        ),
    ]
    industry_group: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            None,
            description="Industry group within the sector, providing a more specific categorization",
        ),
    ]
    industry: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            None,
            description="Detailed industry categorization for the company, like 'Real Estate Management & Development'",
        ),
    ]
    website: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(default=None, description="URL of the company's official website"),
    ]
    market_capitalization: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Market capitalization value",
        ),
    ]

    book_value: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Represents the net value of a company as recorded on "
            "its balance sheet. It reflects the amount shareholders "
            "would theoretically receive if all the company's assets "
            "were sold and all its liabilities were paid off",
        ),
    ]
    price_to_book: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Shares",
        ),
    ]
    trailing_price_to_earnings: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Calculated using a company's earnings over the last 12 months",
        ),
    ]

    dividend_yield: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Calculated using projected earnings for the next 12 months",
        ),
    ]
    dividend_rate: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Calculated using projected earnings for the next 12 months",
        ),
    ]
    trailing_earnings_per_share: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="EPS (Earnings Per Share) is a measure of a company's "
            "profitability on a per-share basis, often used by investors to gauge financial performance",
        ),
    ]
    forward_earnings_per_share: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="EPS (Earnings Per Share) is a measure of a company's "
            "profitability on a per-share basis, often used by investors to gauge financial performance",
        ),
    ]
    return_on_equity: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Indicates how efficiently a company uses shareholder equity to generate profits.",
        ),
    ]
    operating_margins: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Percentage of revenue remaining after operating expenses",
        ),
    ]
    gross_margins: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Percentage of revenue remaining after covering the cost of goods sold (COGS)",
        ),
    ]
    revenue_growth: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="The percentage increase in revenue over a specific period",
        ),
    ]


class Equity(BaseEquity):

    state: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            default=None,
            description="State or province where the company's headquarters is located, if applicable",
        ),
    ]
    city: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(None, description="City where the company's headquarters is located"),
    ]
    zipcode: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(default=None, description="Postal code for the company's headquarters"),
    ]

    market_cap: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            default=None,
            description="Market capitalization category, such as 'Large Cap' or 'Small Cap'",
        ),
    ]

    shares_outstanding: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="The total number of shares that a company has issued to investors, "
            "including those held by insiders",
        ),
    ]
    float_shares: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="The portion of Shares Outstanding that is freely available "
            "for public trading on the open market",
        ),
    ]
    short_shares: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Refers to the number of shares of a company that have been sold "
            "short by investors but not yet repurchased or covered.",
        ),
    ]

    forward_price_to_earnings: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Calculated using projected earnings for the next 12 months",
        ),
    ]
    revenue_per_share: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Indicates how much revenue a company generates for each outstanding share of its stock",
        ),
    ]
    quick_ratio: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="The Quick Ratio measures a company's ability "
            "to pay its short-term liabilities using its most liquid assets, excluding inventory",
        ),
    ]
    current_ratio: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="The Current Ratio measures a company's ability "
            "to pay its short-term liabilities (due within a year) with its total current assets",
        ),
    ]
    earning_growth: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="The rate at which a company's net income increases.",
        ),
    ]
    trailing_peg_ratio: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Combines the P/E ratio with earnings growth to assess valuation in the context of growth",
        ),
    ]
    trailing_price_to_sales: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Measures the stock price relative to revenue per share",
        ),
    ]
    return_on_assets: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Measures how effectively a company uses its assets to generate profit",
        ),
    ]
    short_ratio: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="It provides insight into how heavily shorted "
            "a stock is and the potential for a short squeeze.",
        ),
    ]
    timezone: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            default=None,
            description="The timezone of the exchange where the stock is traded",
        ),
    ]

    cusip: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            default=None,
            description="CUSIP identifier for the company's stock (mainly used in the US)",
        ),
    ]
    figi: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            default=None,
            description="Financial Instrument Global Identifier (FIGI) for the company",
        ),
    ]
    composite_figi: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            default=None,
            description="Composite FIGI, a global identifier that aggregates multiple instruments",
        ),
    ]
    shareclass_figi: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            default=None,
            description="FIGI specific to a share class, distinguishing between types of shares",
        ),
    ]



from datetime import datetime
from typing import Optional, List, Dict, Any, cast, ClassVar

import requests  # type: ignore

from bearish.exchanges.exchanges import Countries
from bearish.models.assets.assets import Assets
from bearish.models.base import SourceBase
from bearish.models.financials.base import Financials
from bearish.models.price.price import Price
from bearish.models.query.query import AssetQuery
from bearish.sources.base import AbstractSource
from bearish.types import Sources, SeriesLength
from bearish.utils.utils import get_start_date


def compute_url(ticker: str, api_key: str, from_: Optional[str] = None) -> str:
    from_ = from_ if from_ else "2019-01-02"
    return f"https://api.tiingo.com/tiingo/daily/{ticker.lower()}/prices?startDate={from_}&token={api_key}"


def read_api(
    api_key: str, ticker: str, from_: Optional[str] = None
) -> List[Dict[str, Any]]:
    headers = {"Content-Type": "application/json"}
    request_response = requests.get(
        compute_url(ticker, api_key, from_), headers=headers, timeout=10
    )
    return cast(List[Dict[str, Any]], request_response.json())


class TiingoSourceBase(SourceBase):
    __source__: Sources = "Tiingo"


class TiingoPrice(TiingoSourceBase, Price):
    __alias__: ClassVar[Dict[str, str]] = {
        "date": "date",
        "close": "close",
        "high": "high",
        "low": "low",
        "open": "open",
        "volume": "volume",
        "divCash": "dividends",
        "splitFactor": "stock_splits",
    }


class TiingoSource(TiingoSourceBase, AbstractSource):
    countries: List[Countries] = ["US"]  # noqa: RUF012

    def set_api_key(self, api_key: str) -> None:
        TiingoSourceBase.__api_key__ = api_key

    def _read_assets(self, query: Optional[AssetQuery] = None) -> Assets:
        return Assets()

    def _read_financials(self, tickers: List[str]) -> List[Financials]:
        return [Financials()]

    def _read_series(self, tickers: List[str], type: SeriesLength) -> List[TiingoPrice]:  # type: ignore
        from_ = get_start_date(type)
        prices = []
        for ticker in tickers:
            datas = read_api(self.__api_key__, ticker, from_)
            prices.extend(
                [
                    TiingoPrice.model_validate(
                        {
                            **data,
                            "symbol": ticker,
                            "date": datetime.strptime(
                                data["date"], "%Y-%m-%dT%H:%M:%S.%fZ"
                            ).date(),
                        }
                    )
                    for data in datas
                ]
            )
        return prices



import logging
from datetime import datetime, date
from functools import cached_property
from pathlib import Path
from typing import List, TYPE_CHECKING, Type, Union, Any, Optional

import pandas as pd
from pydantic import BaseModel, ConfigDict
from sqlalchemy import Engine, create_engine, insert
from sqlmodel import Session, select
from sqlmodel.main import SQLModel


from bearish.database.schemas import (
    EquityORM,
    CurrencyORM,
    CryptoORM,
    EtfORM,
    FinancialMetricsORM,
    CashFlowORM,
    BalanceSheetORM,
    PriceORM,
    SourcesORM,
    EarningsDateORM,
    QuarterlyFinancialMetricsORM,
    QuarterlyCashFlowORM,
    QuarterlyBalanceSheetORM,
    PriceTrackerORM,
    FinancialsTrackerORM,
    IndexORM,
    SecORM,
)
from bearish.database.scripts.upgrade import upgrade
from bearish.exchanges.exchanges import ExchangeQuery
from bearish.interface.interface import BearishDbBase
from bearish.models.assets.assets import Assets
from bearish.models.base import (
    TrackerQuery,
    Ticker,
    BaseTracker,
    FinancialsTracker,
    PriceTracker,
)
from bearish.models.financials.balance_sheet import BalanceSheet, QuarterlyBalanceSheet
from bearish.models.financials.base import Financials, ManyFinancials
from bearish.models.financials.cash_flow import CashFlow, QuarterlyCashFlow
from bearish.models.financials.earnings_date import EarningsDate
from bearish.models.financials.metrics import (
    FinancialMetrics,
    QuarterlyFinancialMetrics,
)
from bearish.models.price.price import Price
from bearish.models.sec.sec import Sec
from bearish.utils.utils import batch

if TYPE_CHECKING:
    from bearish.models.query.query import AssetQuery

logger = logging.getLogger(__name__)

BATCH_SIZE = 5000


class BearishDb(BearishDbBase):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    database_path: Path
    auto_migration: bool = True

    @cached_property
    def _engine(self) -> Engine:
        database_url = f"sqlite:///{Path(self.database_path)}"
        if self.auto_migration:
            upgrade(database_url)
        engine = create_engine(database_url)
        return engine

    def model_post_init(self, __context: Any) -> None:
        self._engine  # noqa: B018

    def _write_assets(self, assets: Assets) -> None:
        with Session(self._engine) as session:
            objects_orm = (
                [EquityORM(**object.model_dump()) for object in assets.equities]
                + [CurrencyORM(**object.model_dump()) for object in assets.currencies]
                + [CryptoORM(**object.model_dump()) for object in assets.cryptos]
                + [EtfORM(**object.model_dump()) for object in assets.etfs]
                + [IndexORM(**object.model_dump()) for object in assets.index]
            )
            logger.debug(
                f"writing assets to database. Number of assets: {len(objects_orm)}"
            )

            session.add_all(objects_orm)
            session.commit()

    def _write_series(
        self, series: List["Price"], table: Optional[Type[SQLModel]] = None
    ) -> None:
        price_orm = table or PriceORM
        with Session(self._engine) as session:
            data = [serie.model_dump() for serie in series]
            chunks = batch(data, BATCH_SIZE)
            for chunk in chunks:
                stmt = insert(price_orm).prefix_with("OR REPLACE").values(chunk)
                session.exec(stmt)  # type: ignore
            session.commit()

    def _write_sec(self, secs: List["Sec"]) -> None:

        with Session(self._engine) as session:
            data = [sec.model_dump() for sec in secs]
            chunks = batch(data, BATCH_SIZE)
            for chunk in chunks:
                stmt = insert(SecORM).prefix_with("OR REPLACE").values(chunk)
                session.exec(stmt)  # type: ignore
            session.commit()

    def _read_sec(self, ticker: str) -> List[Sec]:
        with Session(self._engine) as session:
            stmt = select(SecORM).where(SecORM.ticker == ticker)
            sources = session.exec(stmt).all()
            return [Sec.model_validate(source.model_dump()) for source in sources]

    def _write_financials(self, financials: List[Financials]) -> None:
        many_financials = ManyFinancials(financials=financials)
        self._write_financials_series(
            many_financials.get("financial_metrics"), FinancialMetricsORM  # type: ignore
        )
        self._write_financials_series(many_financials.get("cash_flows"), CashFlowORM)  # type: ignore
        self._write_financials_series(
            many_financials.get("balance_sheets"), BalanceSheetORM  # type: ignore
        )
        self._write_financials_series(
            many_financials.get("earnings_date"), EarningsDateORM  # type: ignore
        )
        self._write_financials_series(
            many_financials.get("quarterly_financial_metrics"),  # type: ignore
            QuarterlyFinancialMetricsORM,
        )
        self._write_financials_series(
            many_financials.get("quarterly_cash_flows"), QuarterlyCashFlowORM  # type: ignore
        )
        self._write_financials_series(
            many_financials.get("quarterly_balance_sheets"), QuarterlyBalanceSheetORM  # type: ignore
        )

    def _write_financials_series(
        self,
        series: Union[
            List[CashFlow],
            List[FinancialMetrics],
            List[BalanceSheet],
            List[EarningsDate],
            List[QuarterlyCashFlow],
            List[QuarterlyFinancialMetrics],
            List[QuarterlyBalanceSheet],
        ],
        table: Type[SQLModel],
    ) -> None:
        if not series:
            logger.warning(f"No data found for '{[serie.symbol for serie in series]}'")
            return None
        with Session(self._engine) as session:
            data = [serie.model_dump() for serie in series]
            chunks = batch(data, BATCH_SIZE)
            for chunk in chunks:
                stmt = insert(table).prefix_with("OR REPLACE").values(chunk)
                session.exec(stmt)  # type: ignore
            session.commit()

    def _read_series(
        self,
        query: "AssetQuery",
        months: int = 1,
        table: Optional[Type[SQLModel]] = None,
    ) -> List[Price]:
        end_date = datetime.now()
        start_date = end_date - pd.Timedelta(days=months * 31)
        table = table or PriceORM
        with Session(self._engine) as session:
            query_ = select(table)
            query_ = query_.where(table.symbol.in_(query.symbols.all())).where(  # type: ignore
                table.date.between(start_date, end_date)  # type: ignore
            )
            series = session.exec(query_).all()
            return [Price.model_validate(serie) for serie in series]

    def _read_financials(self, query: "AssetQuery") -> Financials:
        with Session(self._engine) as session:
            financial_metrics = self._read_asset_type(
                session, FinancialMetrics, FinancialMetricsORM, query
            )
            cash_flows = self._read_asset_type(session, CashFlow, CashFlowORM, query)
            balance_sheets = self._read_asset_type(
                session, BalanceSheet, BalanceSheetORM, query
            )
            earnings_date = self._read_asset_type(
                session, EarningsDate, EarningsDateORM, query
            )
            quarterly_financial_metrics = self._read_asset_type(
                session, QuarterlyFinancialMetrics, QuarterlyFinancialMetricsORM, query
            )
            quarterly_cash_flows = self._read_asset_type(
                session, QuarterlyCashFlow, QuarterlyCashFlowORM, query
            )
            quarterly_balance_sheets = self._read_asset_type(
                session, QuarterlyBalanceSheet, QuarterlyBalanceSheetORM, query
            )
            return Financials(
                financial_metrics=financial_metrics,
                cash_flows=cash_flows,
                balance_sheets=balance_sheets,
                quarterly_financial_metrics=quarterly_financial_metrics,
                quarterly_cash_flows=quarterly_cash_flows,
                quarterly_balance_sheets=quarterly_balance_sheets,
                earnings_date=earnings_date,
            )

    def _read_assets(self, query: "AssetQuery") -> Assets:
        with Session(self._engine) as session:
            from bearish.models.assets.equity import Equity
            from bearish.models.assets.crypto import Crypto
            from bearish.models.assets.currency import Currency
            from bearish.models.assets.etfs import Etf
            from bearish.models.assets.index import Index

            equities = self._read_asset_type(session, Equity, EquityORM, query)
            currencies = self._read_asset_type(session, Currency, CurrencyORM, query)
            cryptos = self._read_asset_type(session, Crypto, CryptoORM, query)
            etfs = self._read_asset_type(session, Etf, EtfORM, query)
            index = self._read_asset_type(session, Index, IndexORM, query)
            return Assets(
                equities=equities,
                currencies=currencies,
                cryptos=cryptos,
                etfs=etfs,
                index=index,
            )

    def _read_asset_type(
        self,
        session: Session,
        table: Type[BaseModel],
        orm_table: Type[
            Union[
                EquityORM,
                CurrencyORM,
                CryptoORM,
                EtfORM,
                BalanceSheetORM,
                CashFlowORM,
                FinancialMetricsORM,
                QuarterlyBalanceSheetORM,
                QuarterlyCashFlowORM,
                QuarterlyFinancialMetricsORM,
                EarningsDateORM,
                IndexORM,
            ]
        ],
        query: "AssetQuery",
    ) -> List[BaseModel]:
        if query.symbols.all():
            query_ = select(orm_table).where(orm_table.symbol.in_(query.symbols.all()))  # type: ignore
        else:
            query_ = select(orm_table)
            if query.countries:
                query_ = query_.where(orm_table.country.in_(query.countries))  # type: ignore
            if query.exchanges:
                query_ = query_.where(orm_table.exchange.in_(query.exchanges))  # type: ignore
        if query.excluded_sources:
            query_ = query_.where(~orm_table.source.in_(query.excluded_sources))  # type: ignore

        assets = session.exec(query_).all()
        return [table.model_validate(asset) for asset in assets]

    def _read_sources(self) -> List[str]:
        with Session(self._engine) as session:
            query_ = select(SourcesORM).distinct()
            sources = session.exec(query_).all()
            return [source.source for source in sources]

    def _write_source(self, source: str) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(SourcesORM)
                .prefix_with("OR REPLACE")
                .values([{"source": source}])
            )

            session.exec(stmt)  # type: ignore
            session.commit()

    def _write_trackers(
        self,
        trackers: List[FinancialsTracker] | List[PriceTracker],
        tracker_type: Type[BaseTracker],
    ) -> None:
        with Session(self._engine) as session:
            orm_class = (
                PriceTrackerORM
                if tracker_type is PriceTracker
                else FinancialsTrackerORM
            )
            stmt = (
                insert(orm_class)
                .prefix_with("OR REPLACE")
                .values([t.model_dump() for t in trackers])
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def _read_tracker(
        self,
        tracker_query: TrackerQuery,
        tracker_type: Union[Type[PriceTracker], Type[FinancialsTracker]],
    ) -> List[Ticker]:
        with Session(self._engine) as session:
            tracker_orm = (
                PriceTrackerORM
                if tracker_type is PriceTracker
                else FinancialsTrackerORM
            )
            query = select(tracker_orm.symbol, tracker_orm.exchange, tracker_orm.source)
            if tracker_query.exchange:
                query = query.where(tracker_orm.exchange == tracker_query.exchange)

            tracker_orm = session.exec(query).all()  # type: ignore
            return [
                Ticker(symbol=t[0], exchange=t[1], source=t[2]) for t in tracker_orm  # type: ignore
            ]

    def _get_tickers(self, exchange_query: ExchangeQuery) -> List[Ticker]:
        if not exchange_query.sources:
            query = f"""SELECT symbol, exchange from equity where {exchange_query.to_suffixes_sql_statement()} 
                OR exchange IN {exchange_query.to_aliases_sql_statement()};"""
        else:
            query = f"""SELECT symbol, exchange from equity where source IN {exchange_query.to_sources_sql_statement()} 
            AND ( {exchange_query.to_suffixes_sql_statement()} 
                 OR exchange IN {exchange_query.to_aliases_sql_statement()});"""

        symbols = pd.read_sql(
            query,
            con=self._engine,
        )
        return [
            Ticker(symbol=symbol["symbol"], exchange=symbol["exchange"])
            for symbol in symbols.to_dict(orient="records")
        ]

    def _read_query(self, query: str) -> pd.DataFrame:
        return pd.read_sql(
            query,
            con=self._engine,
        )

    def read_price_tracker(self, symbol: str) -> Optional[date]:
        with Session(self._engine) as session:
            query = select(PriceTrackerORM.date).where(PriceTrackerORM.symbol == symbol)  # type: ignore
            result = session.exec(query).first()
            if result is None:
                return None
            return result  # type: ignore



from typing import Annotated, Optional

from pydantic import BeforeValidator, Field

from bearish.models.base import DataSourceBase
from bearish.utils.utils import to_float


class BalanceSheet(DataSourceBase):
    treasury_stock: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Shares held in the company's treasury"),
    ]

    common_stock_shares_outstanding: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Number of common shares outstanding"),
    ]

    common_stock: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Value of common stock issued"),
    ]

    short_long_term_debt_total: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Total short and long-term debt"),
    ]

    capital_lease_obligations: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Obligations under capital leases"),
    ]

    total_shareholder_equity: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Total equity held by shareholders"),
    ]

    retained_earnings: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Accumulated retained earnings"),
    ]

    total_liabilities: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Total liabilities, net of minority interest"),
    ]

    total_non_current_liabilities: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Total non-current liabilities"),
    ]

    other_non_current_liabilities: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Other non-current liabilities"),
    ]

    long_term_debt: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Long-term debt obligations"),
    ]

    total_current_liabilities: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Total current liabilities"),
    ]

    other_current_liabilities: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Other current liabilities"),
    ]

    current_debt: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Current debt obligations"),
    ]

    current_accounts_payable: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Accounts payable amount"),
    ]

    total_assets: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Total assets held by the company"),
    ]

    total_non_current_assets: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Total non-current assets"),
    ]

    other_non_current_assets: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Other non-current assets"),
    ]

    accumulated_depreciation_amortization_ppe: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Accumulated depreciation and amortization for property, plant, and equipment",
        ),
    ]

    property_plant_equipment: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Gross property, plant, and equipment"),
    ]

    total_current_assets: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Total current assets"),
    ]

    other_current_assets: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Other current assets"),
    ]

    inventory: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Inventory held by the company"),
    ]

    current_net_receivables: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Net receivables amount"),
    ]

    cash_and_short_term_investments: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Cash and short-term investments"),
    ]

    cash_and_cash_equivalents_at_carrying_value: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Cash and cash equivalents at carrying value"),
    ]


class QuarterlyBalanceSheet(BalanceSheet): ...



"""price growth

Revision ID: b62d0afd2675
Revises: 5f6984590216
Create Date: 2025-06-15 07:06:37.914668

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "b62d0afd2675"
down_revision: Union[str, None] = "5f6984590216"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_analysis_source"))
        batch_op.drop_index(batch_op.f("ix_analysis_symbol"))

    op.drop_table("analysis")
    with op.batch_alter_table("view", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_view_source"))
        batch_op.drop_index(batch_op.f("ix_view_symbol"))

    op.drop_table("view")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "view",
        sa.Column("symbol", sa.VARCHAR(), nullable=False),
        sa.Column("exchange", sa.VARCHAR(), nullable=True),
        sa.Column("source", sa.VARCHAR(), nullable=False),
        sa.Column("date", sa.DATE(), nullable=False),
        sa.Column("created_at", sa.DATE(), nullable=False),
        sa.Column("name", sa.VARCHAR(), nullable=True),
        sa.Column("isin", sa.VARCHAR(), nullable=True),
        sa.Column("view_name", sa.VARCHAR(), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("view", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_view_symbol"), ["symbol"], unique=False)
        batch_op.create_index(batch_op.f("ix_view_source"), ["source"], unique=False)

    op.create_table(
        "analysis",
        sa.Column("quarterly_positive_free_cash_flow", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_growing_operating_cash_flow", sa.FLOAT(), nullable=True),
        sa.Column(
            "quarterly_operating_cash_flow_is_higher_than_net_income",
            sa.FLOAT(),
            nullable=True,
        ),
        sa.Column("quarterly_mean_capex_ratio", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_max_capex_ratio", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_min_capex_ratio", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_mean_dividend_payout_ratio", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_max_dividend_payout_ratio", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_min_dividend_payout_ratio", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_positive_net_income", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_positive_operating_income", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_growing_net_income", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_growing_operating_income", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_positive_diluted_eps", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_positive_basic_eps", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_growing_basic_eps", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_growing_diluted_eps", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_positive_debt_to_equity", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_positive_return_on_assets", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_positive_return_on_equity", sa.FLOAT(), nullable=True),
        sa.Column("quarterly_earning_per_share", sa.FLOAT(), nullable=True),
        sa.Column("positive_free_cash_flow", sa.FLOAT(), nullable=True),
        sa.Column("growing_operating_cash_flow", sa.FLOAT(), nullable=True),
        sa.Column(
            "operating_cash_flow_is_higher_than_net_income", sa.FLOAT(), nullable=True
        ),
        sa.Column("mean_capex_ratio", sa.FLOAT(), nullable=True),
        sa.Column("max_capex_ratio", sa.FLOAT(), nullable=True),
        sa.Column("min_capex_ratio", sa.FLOAT(), nullable=True),
        sa.Column("mean_dividend_payout_ratio", sa.FLOAT(), nullable=True),
        sa.Column("max_dividend_payout_ratio", sa.FLOAT(), nullable=True),
        sa.Column("min_dividend_payout_ratio", sa.FLOAT(), nullable=True),
        sa.Column("positive_net_income", sa.FLOAT(), nullable=True),
        sa.Column("positive_operating_income", sa.FLOAT(), nullable=True),
        sa.Column("growing_net_income", sa.FLOAT(), nullable=True),
        sa.Column("growing_operating_income", sa.FLOAT(), nullable=True),
        sa.Column("positive_diluted_eps", sa.FLOAT(), nullable=True),
        sa.Column("positive_basic_eps", sa.FLOAT(), nullable=True),
        sa.Column("growing_basic_eps", sa.FLOAT(), nullable=True),
        sa.Column("growing_diluted_eps", sa.FLOAT(), nullable=True),
        sa.Column("positive_debt_to_equity", sa.FLOAT(), nullable=True),
        sa.Column("positive_return_on_assets", sa.FLOAT(), nullable=True),
        sa.Column("positive_return_on_equity", sa.FLOAT(), nullable=True),
        sa.Column("earning_per_share", sa.FLOAT(), nullable=True),
        sa.Column("rsi_last_value", sa.FLOAT(), nullable=True),
        sa.Column("macd_12_26_9_buy_date", sa.DATE(), nullable=True),
        sa.Column("ma_50_200_buy_date", sa.DATE(), nullable=True),
        sa.Column("slope_7", sa.FLOAT(), nullable=True),
        sa.Column("slope_14", sa.FLOAT(), nullable=True),
        sa.Column("slope_30", sa.FLOAT(), nullable=True),
        sa.Column("slope_60", sa.FLOAT(), nullable=True),
        sa.Column("last_adx", sa.FLOAT(), nullable=True),
        sa.Column("last_dmp", sa.FLOAT(), nullable=True),
        sa.Column("last_dmn", sa.FLOAT(), nullable=True),
        sa.Column("last_price", sa.FLOAT(), nullable=True),
        sa.Column("last_price_date", sa.DATE(), nullable=True),
        sa.Column("year_to_date_growth", sa.FLOAT(), nullable=True),
        sa.Column("last_52_weeks_growth", sa.FLOAT(), nullable=True),
        sa.Column("last_week_growth", sa.FLOAT(), nullable=True),
        sa.Column("last_month_growth", sa.FLOAT(), nullable=True),
        sa.Column("last_year_growth", sa.FLOAT(), nullable=True),
        sa.Column("year_to_date_max_growth", sa.FLOAT(), nullable=True),
        sa.Column("last_week_max_growth", sa.FLOAT(), nullable=True),
        sa.Column("last_month_max_growth", sa.FLOAT(), nullable=True),
        sa.Column("last_year_max_growth", sa.FLOAT(), nullable=True),
        sa.Column("macd_12_26_9_buy", sa.FLOAT(), nullable=True),
        sa.Column("star_yoy", sa.FLOAT(), nullable=True),
        sa.Column("star_wow", sa.FLOAT(), nullable=True),
        sa.Column("star_mom", sa.FLOAT(), nullable=True),
        sa.Column("symbol", sa.VARCHAR(), nullable=False),
        sa.Column("exchange", sa.VARCHAR(), nullable=True),
        sa.Column("source", sa.VARCHAR(), nullable=False),
        sa.Column("date", sa.DATE(), nullable=False),
        sa.Column("created_at", sa.DATE(), nullable=False),
        sa.Column("name", sa.VARCHAR(), nullable=True),
        sa.Column("isin", sa.VARCHAR(), nullable=True),
        sa.Column("base_symbol", sa.VARCHAR(), nullable=False),
        sa.Column("modifier", sa.VARCHAR(), nullable=True),
        sa.Column("summary", sa.VARCHAR(), nullable=True),
        sa.Column("currency", sa.VARCHAR(), nullable=True),
        sa.Column("market", sa.VARCHAR(), nullable=True),
        sa.Column("country", sa.VARCHAR(), nullable=True),
        sa.Column("sector", sa.VARCHAR(), nullable=True),
        sa.Column("industry_group", sa.VARCHAR(), nullable=True),
        sa.Column("industry", sa.VARCHAR(), nullable=True),
        sa.Column("website", sa.VARCHAR(), nullable=True),
        sa.Column("market_capitalization", sa.FLOAT(), nullable=True),
        sa.Column("book_value", sa.FLOAT(), nullable=True),
        sa.Column("price_to_book", sa.FLOAT(), nullable=True),
        sa.Column("trailing_price_to_earnings", sa.FLOAT(), nullable=True),
        sa.Column("dividend_yield", sa.FLOAT(), nullable=True),
        sa.Column("dividend_rate", sa.FLOAT(), nullable=True),
        sa.Column("trailing_earnings_per_share", sa.FLOAT(), nullable=True),
        sa.Column("forward_earnings_per_share", sa.FLOAT(), nullable=True),
        sa.Column("return_on_equity", sa.FLOAT(), nullable=True),
        sa.Column("operating_margins", sa.FLOAT(), nullable=True),
        sa.Column("gross_margins", sa.FLOAT(), nullable=True),
        sa.Column("revenue_growth", sa.FLOAT(), nullable=True),
        sa.Column("price_per_earning_ratio", sa.FLOAT(), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_analysis_symbol"), ["symbol"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_analysis_source"), ["source"], unique=False
        )

    # ### end Alembic commands ###



from typing import Annotated, Optional

from pydantic import BeforeValidator, Field

from bearish.models.base import DataSourceBase
from bearish.utils.utils import to_float


class EarningsDate(DataSourceBase):
    eps_estimate: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Estimated EPS"),
    ]
    eps_reported: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Reported EPS"),
    ]
    revenue_estimate: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Estimated Revenue"),
    ]
    revenue_reported: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Reported Revenue"),
    ]



import abc
import logging
from datetime import date
from pathlib import Path
from typing import List, Type, Union, Optional

import pandas as pd
from pydantic import BaseModel, ConfigDict, validate_call
from sqlmodel import SQLModel

from bearish.exchanges.exchanges import ExchangeQuery
from bearish.models.assets.assets import Assets
from bearish.models.base import (
    TrackerQuery,
    Ticker,
    PriceTracker,
    FinancialsTracker,
    BaseTracker,
)
from bearish.models.financials.base import Financials
from bearish.models.financials.earnings_date import EarningsDate
from bearish.models.price.price import Price
from bearish.models.query.query import AssetQuery
from bearish.models.sec.sec import Sec
from bearish.utils.utils import observability


logger = logging.getLogger(__name__)


class BearishDbBase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    database_path: Path

    @validate_call
    def write_assets(self, assets: Assets) -> None:
        return self._write_assets(assets)

    @observability
    @validate_call
    def write_series(
        self, series: List[Price], table: Optional[Type[SQLModel]] = None
    ) -> None:
        return self._write_series(series, table=table)

    @validate_call
    def write_sec(self, secs: List[Sec]) -> None:
        return self._write_sec(secs)

    @validate_call
    def read_sec(self, ticker: str) -> List[Sec]:
        return self._read_sec(ticker)

    @observability
    @validate_call
    def write_financials(self, financials: List[Financials]) -> None:
        return self._write_financials(financials)

    @validate_call
    def read_series(
        self, query: AssetQuery, months: int = 1, table: Optional[Type[SQLModel]] = None
    ) -> List[Price]:
        return self._read_series(query, months, table=table)

    @validate_call
    def read_financials(self, query: AssetQuery) -> Financials:
        return self._read_financials(query)

    def read_earnings_date(self, query: AssetQuery) -> List[EarningsDate]:
        financials = self.read_financials(query)
        return financials.earnings_date

    @validate_call
    def read_assets(self, query: AssetQuery) -> Assets:
        return self._read_assets(query)

    @validate_call
    def read_sources(self) -> List[str]:
        return self._read_sources()

    @validate_call
    def get_tickers(self, exchange_query: ExchangeQuery) -> List[Ticker]:
        return list(set(self._get_tickers(exchange_query)))

    def write_source(self, source: str) -> None:
        return self._write_source(source)

    @validate_call
    def read_tracker(
        self,
        tracker_query: TrackerQuery,
        tracker_type: Union[Type[PriceTracker], Type[FinancialsTracker]],
    ) -> List[Ticker]:
        return self._read_tracker(tracker_query, tracker_type)

    def write_trackers(
        self, trackers: List[FinancialsTracker] | List[PriceTracker]
    ) -> None:
        tracker_type = type(trackers[0])
        return self._write_trackers(trackers, tracker_type)

    def read_query(self, query: str) -> pd.DataFrame:
        return self._read_query(query)

    @abc.abstractmethod
    def _write_assets(self, assets: Assets) -> None: ...

    @abc.abstractmethod
    def _write_series(
        self, series: List[Price], table: Optional[Type[SQLModel]] = None
    ) -> None: ...

    @abc.abstractmethod
    def _write_sec(self, secs: List[Sec]) -> None: ...

    @abc.abstractmethod
    def _read_sec(self, ticker: str) -> List[Sec]: ...

    @abc.abstractmethod
    def _write_financials(self, financials: List[Financials]) -> None: ...

    @abc.abstractmethod
    def _read_series(
        self, query: AssetQuery, months: int = 1, table: Optional[Type[SQLModel]] = None
    ) -> List[Price]: ...

    @abc.abstractmethod
    def _read_financials(self, query: AssetQuery) -> Financials: ...

    @abc.abstractmethod
    def _read_assets(self, query: AssetQuery) -> Assets: ...

    @abc.abstractmethod
    def _write_source(self, source: str) -> None: ...

    @abc.abstractmethod
    def _read_sources(self) -> List[str]: ...

    @abc.abstractmethod
    def _read_tracker(
        self,
        tracker_query: TrackerQuery,
        tracker_type: Union[Type[PriceTracker], Type[FinancialsTracker]],
    ) -> List[Ticker]: ...
    @abc.abstractmethod
    def _write_trackers(
        self,
        trackers: List[PriceTracker] | List[FinancialsTracker],
        tracker_type: Type[BaseTracker],
    ) -> None: ...

    @abc.abstractmethod
    def _get_tickers(self, exchange_query: ExchangeQuery) -> List[Ticker]: ...

    @abc.abstractmethod
    def _read_query(self, query: str) -> pd.DataFrame: ...

    @abc.abstractmethod
    def read_price_tracker(self, symbol: str) -> Optional[date]: ...
